//
// Rational.java
//
// Bill Seymour, 2024-12-25
//
// Copyright Bill Seymour 2024.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This is an open-source rational number in Java that uses
// BigIntegers for its numerator and denominator.
//

import java.math.*;

package rat;

public final class Rational extends Number implements Comparable<Rational>
{
    private BigInteger num, den;

    public static final Rational ZERO = new Rational();
    public static final Rational ONE  = new Rational(BigInteger.ONE);
    public static final Rational HALF = new Rational(1L, 2L);

    public Rational()
    {
        num = BigInteger.ZERO;
        den = BigInteger.ONE;
    }
    public Rational(long val)
    {
        num = BigInteger.valueOf(val);
        den = BigInteger.ONE;
    }
    public Rational(BigInteger val)
    {
        num = val;
        den = BigInteger.ONE;
    }

    //
    // Euclid's algorithm for greatest common divisor
    // (Knuth, vol. 2, sect. 4.5.2):
    //
    // Constraint:  a and c non-negative and not both zero
    //
    private static long gcd(long a, long c)
    {
        for (long r; c != 0; )
        {
            r = a % c;
            a = c;
            c = r;
        }
        return a;
    }
    public Rational(long n, long d)
    {
        if (0 == d)
        {
            throw new ArithmeticException(
                "Constructing Rational with zero denominator");
        }
        if (0 == n)
        {
            num = BigInteger.ZERO;
            den = BigInteger.ONE;
            return; // NB:  early return
        }
        if (d < 0)
        {
            n = -n;
            d = -d;
        }
        if (d != 1)
        {
            long cd = gcd(Math.abs(n), d);
            if (cd > 1)
            {
                n /= cd;
                d /= cd;
            }
        }
        num = BigInteger.valueOf(n);
        den = BigInteger.valueOf(d);
    }

    //
    // The following helper normalizes the fraction such that the denominator
    // is greater than zero and the fraction is reduced to lowest terms.
    // It's shared by Rational(BigInteger,BigInteger) and Rational(String,int).
    //
    private BigInteger[] norm(BigInteger n, BigInteger d)
    {
        int dsign = d.signum();
        if (0 == dsign)
        {
            throw new ArithmeticException(
                "Constructing Rational with zero denominator");
        }
        if (dsign < 0)
        {
            n = n.negate();
            d = d.negate();
        }
        if (n.equals(BigInteger.ZERO))
        {
            d = BigInteger.ONE;
        }
        else if (!d.equals(BigInteger.ONE))
        {
            BigInteger cd = d.gcd(n);
            if (cd.compareTo(BigInteger.ONE) > 0)
            {
                n = n.divide(cd);
                d = d.divide(cd);
            }
        }
        return new BigInteger[2] { n, d };
    }
    public Rational(BigInteger n, BigInteger d)
    {
        BigInteger[] both = norm(n, d);
        num = both[0];
        den = both[1];
    }

    public Rational(double val, double err)
    {
        if (!Double.isFinite(val))
        {
            throw new IllegalArgumentException(
                "Can't construct Rational from non-finite double");
        }

        boolean isNeg = val < 0.0;
        if (isNeg)
        {
            val = -val;
        }

        if (0.0 == val) // probably premature optimization but good for sanity
        {
            num = BigInteger.ZERO;
            den = BigInteger.ONE;
            return; // NB:  early return
        }

        // continued fractions (Knuth, vol. 2, sect. 4.5.3):

        double maxError = Math.abs(val * err);
        if (!Double.isFinite(maxError) || maxError < Double.MIN_NORMAL)
        {
            throw new ArithmeticException(
                "Non-finite or extremely small maximum error " +
                "when constructing Rational from double");
        }

        double val0 = val; // current fraction term

        BigInteger int0 = BigInteger.valueOf((long)val0);

        BigInteger num0 = int0,           // current approximation
                   den0 = BigInteger.ONE; // first guess:  int(val) / 1

        BigInteger num1 = BigInteger.ONE, num2;  // prev. and 2nd-prev. numer.
        BigInteger den1 = BigInteger.ZERO, den2; // prev. and 2nd-prev. denom.

        double err0 = Math.abs(val0 - int0.doubleValue());

        while (err0 > maxError)
        {
            val0 = 1.0 / (val0 - int0.doubleValue());
            if (!Double.isFinite(val0))
            {
                break; // the best we can do
            }
            int0 = BigInteger.valueOf((long)val0);

            num2 = num1;
            num1 = num0;
            num0 = int0.multiply(num1).add(num2);

            den2 = den1;
            den1 = den0;
            den0 = int0.multiply(den1).add(den2);

            long double err1 = err0;
            err0 = Math.abs(val - num0.doubleValue() / den0.doubleValue());

            if (err0 >= err1) // no longer converging
            {
                break; // the best we can do
            }
        }

        //
        // Continued fractions result is known to be in lowest terms already.
        //
        num = isNeg() ? num0.negate() : num0;
        den = den0;
    }

    //
    // After AI help generated by Google search:
    //
    // Explanation given:
    //
    // * doubleToLongBits(): Converts a double value to its bit representation.
    // * Bit Manipulation: Extracts the exponent and significand bits from the
    //   bit representation.
    // * Exponent Calculation: Adjusts the exponent value based on IEEE 754.
    // * Significand Calculation: Constructs the significand by setting the
    //   exponent bits to 0x3fe (representing 0.5) and combining it with the
    //   extracted significand bits.
    // * Return Value: Returns an array containing the significand and exponent.
    //
    // (We might need to take this with a grain of salt.)
    //
    private static double[] frexp(double x)
    {
        long bits = Double.doubleToLongBits(x);
        long exponentBits = (bits & 0x7ff0000000000000L) >> 52;
        long significandBits = bits & 0x000fffffffffffffffL;

        int exponent = (int) exponentBits - 1023;

        if (-1023 == exponent) // Subnormal number
        {
            exponent = -1022;
            while ((significandBits & 0x0010000000000000L) == 0)
            {
                significandBits <<= 1;
                exponent--;
            }
            significandBits &= 0x000fffffffffffffffL;
        }

        double significand =
            Double.longBitsToDouble(0x3fe0000000000000L | significandBits);

        return double[2] { significand, exponent };
    }
    //
    // Exact conversion using the frexp helper above:
    //
    public Rational(double val)
    {
        if (!Double.isFinite(val))
        {
            throw new IllegalArgumentException(
                "Can't construct Rational from non-finite double");
        }

        if (0.0 == val) // probably premature optimization but good for sanity
        {
            num = BigInteger.ZERO;
            den = BigInteger.ONE;
            return; // NB:  early return
        }

        boolean isNeg = val < 0.0;
        if (isNeg)
        {
            val = -val; // for sanity
        }

        double[] parts = frexp(val);

        double mant = parts[0];
        int exp = (int)parts[1];

        while (mant != Math.floor(mant))
        {
            mant *= 2;
            --exp;
        }
        // mant is now an integer value,
        // and exp has been adjusted accordingly.

        BigInteger newnum = (new BigDecimal(mant)).toBigInteger();
        BigInteger newden = BigInteger.ONE;

        if (exp > 0)
        {
            newnum = newnum.shiftLeft(exp);
        }
        else if (exp < 0)
        {
            newden = newden.shiftLeft(-exp);
            if (newnum.and(BigInteger.ONE).equals(BigInteger.ZERO))
            {
                //
                // The denominator is some power of 2 and the numerator
                // is even, so we need to do the lowest terms business.
                //
                BigInteger cd = newden.gcd(newnum);
                if (cd.compareTo(BigInteger.ONE) > 0)
                {
                    newnum = newnum.divide(cd);
                    newden = newden.divide(cd);
                }
            }
        }

        num = isNeg ? newnum.negate() : newnum;
        den = newden;
    }

    public Rational(BigDecimal arg)
    {
        BigDecimal val = arg.stripTrailingZeros();
        int sc = val.scale();

        BigInteger n = val.unscaledValue();
        num = sc >= 0 ? n : n.multiply(BigInteger.TEN.pow(-sc));

        den = sc <= 0 ? BigInteger.ONE : BigInteger.TEN.pow(sc);
    }

    public Rational(String val, int radix)
    {
        if (radix < 2)
        {
            throw new IllegalArgumentException(
                "Invalid radix when constructing a Rational from a String");
        }

        int slash = val.indexOf('/');
        if (-1 == slash)
        {
            int point = val.indexOf('.');
            if (-1 == point)
            {
                num = new BigInteger(val, radix);
                den = BigInteger.ONE;
            }
            else
            {
                StringBuilder raw = new StringBuilder(val.substring(0, point));
                raw.append(val.substring(point + 1));

                BigInteger n = new BigInteger(raw, radix);
                int fracDigs = val.length() - point - 1;
                if (fracDigs < 1)
                {
                    num = n;
                    den = BigInteger.ONE;
                }
                else
                {
                    BigInteger[] both =
                        norm(newnum, BigInteger.valueOf(radix).pow(fracDigs));
                    num = both[0];
                    den = both[1];
                }
            }
        }
        else
        {
            BigInteger[] both =
                norm(new BigInteger(val.substring(0, slash), radix),
                     new BigInteger(val.substring(slash + 1), radix));
            num = both[0];
            den = both[1];
        }
    }
    public Rational(String val)
    {
        this(val, 10);
    }

    @Override public boolean equals(Object o)
    {
        if (o == this)
        {
            return true;
        }
        if (!(o instanceof Rational))
        {
            return false;
        }
        Rational rhs = (Rational)o;
        return this.num.equals(rhs.num) && this.den.equals(rhs.den);
    }
    @Override public int compareTo(Rational rhs)
    {
        return num.multiply(rhs.den).subtract(den.multiply(rhs.num)).signum();
    }
    public int signum()
    {
        return num.signum();
    }

    @Override public int hashCode()
    {
        return num.hashCode() ^ den.hashCode();
    }

    public BigInteger numerator()
    {
        return num;
    }
    public BigInteger denominator()
    {
        return den;
    }

    public String toString(int radix, boolean showden1)
    {
        StringBuilder retval = new StringBuilder(num.toString(radix));
        if (showden1 || !den.equals(BigInteger.ONE))
        {
            retval.append('/');
            retval.append(den.toString(radix));
        }
        return retval;
    }
    public String toString(int radix)
    {
        return toString(radix, false);
    }
    public String toString(boolean showden1)
    {
        return toString(10, showden1);
    }
    public String toString()
    {
        return toString(10, false);
    }

    //
    // We often need to return new Rationals only because the class is
    // advertised to be immutable; but we can break the immutable rule
    // internally if knowledge of that doesn't escape from the method.
    //
    // In the following private helper we know that the result will be
    // in lowest terms, so we can avoid going through the greatest common
    // divisor business which happens in the constructors that take two
    // integer arguments.  This technique is also used in reciprocal(),
    // pow() and sqr().
    //
    private Rational neg()
    {
        Rational newrat = new Rational();
        newrat.num = num.negate();
        newrat.den = den;
        return newrat;
    }
    public Rational negate()
    {
        return num.equals(BigInteger.ZERO) ? this : neg();
    }
    public Rational abs()
    {
        return signum() >= 0 ? this : neg();
    }
    public Rational reciprocal()
    {
        if (num.equals(BigInteger.ZERO))
        {
            throw new ArithmeticException("Reciprocal of Rational zero");
        }
        //
        // We know we're in lowest terms.
        //
        Rational newrat = new Rational();
        newrat.num = den;
        newrat.den = num;
        return newrat;
    }
    public Rational min(Rational other)
    {
        return compareTo(other) < 0 ? this : other;
    }
    public Rational max(Rational other)
    {
        return compareTo(other) >= 0 ? this : other;
    }
    public Rational copySign(Rational other)
    {
        return (signum() < 0) == (other.num.signum() < 0) ? this : negate();
    }

    public Rational add(Rational rhs)
    {
        return new Rational(
            this.num.multiply(rhs.den).add(rhs.num.multiply(this.den)),
            this.den.multiply(rhs.den));
    }
    public Rational subtract(Rational rhs)
    {
        return new Rational(
            this.num.multiply(rhs.den).subtract(rhs.num.multiply(this.den)),
            this.den.multiply(rhs.den));
    }
    public Rational multiply(Rational rhs)
    {
        return new Rational(this.num.multiply(rhs.num),
                            this.den.multiply(rhs.den));
    }
    public Rational divide(Rational rhs)
    {
        if (rhs.num.equals(BigInteger.ZERO)
        {
            throw new ArithmeticException("Rational division by zero");
        }
        // multiply by reciprocal of rhs
        return new Rational(this.num.multiply(rhs.den),
                            this.den.multiply(rhs.num));
    }

    public toBigDecimal(MathContext mc)
    {
        return (new BigDecimal(num, 0)).divide(new BigDecimal(den, 0), mc);
    }
    public toBigDecimal()
    {
        return toBigDecimal(MathContext.DECIMAL128);
    }
    @Override public float floatValue()
    {
        return toBigDecimal().floatValue();
    }
    @Override public double doubleValue()
    {
        return toBigDecimal().doubleValue();
    }

    public Rational round(RoundingMode rnd)
    {
        if (den.equals(BigInteger.ONE)
        {
            return this; // It's already an integer.
        }

        //
        // At this point, we can also infer that the numerator is non-zero
        // and is not divisible by the denominator.
        //
        Rational intPart = new Rational(num.abs().divide(den));
        //
        // And so intPart is an integer (intPart.den == 1) equal to
        // the absolute value of this already rounded toward zero.
        //

        boolean isNeg = this.num.signum() < 0; // We'll change the sign later.

        //
        // Other flags that we might not need to compute:
        //
        boolean isHalf = false;
        boolean isLess = false;
        boolean isEven = false;
        if (RoundingMode.HALF_UP   == rnd ||
            RoundingMode.HALF_DOWN == rnd ||
            RoundingMode.HALF_EVEN == rnd)
        {
            int halfness = this.abs().subtract(intPart).compareTo(HALF);
            isHalf = 0 == halfness;
            isLess = halfness < 0;
            isEven = intPart.num.and(BigInteger.ONE).equals(BigInteger.ZERO);
        }

        switch (rnd)
        {
            case RoundingMode.UP:
                // NB:  intPart is an absolute value
                intPart.num.add(BigInteger.ONE);
                break;
            case RoundingMode.DOWN:
                // already done
                break;

            case RoundingMode.CEILING:
                if (isNeg)
                {
                    intPart.num = intPart.num.subtract(BigInteger.ONE);
                }
                else
                {
                    intPart.num = intPart.num.add(BigInteger.ONE);
                }
                break;
            case RoundingMode.FLOOR:
                if (isNeg)
                {
                    intPart.num = intPart.num.add(BigInteger.ONE);
                }
                else
                {
                    intPart.num = intPart.num.subtract(BigInteger.ONE);
                }
                break;

            case RoundingMode.HALF_UP:
                if (!isLess)
                {
                    intPart.num = intPart.num.add(BigInteger.ONE);
                }
                break;
            case RoundingMode.HALF_DOWN:
                if (isHalf || isLess)
                {
                    intPart.num = intPart.num.subtract(BigInteger.ONE);
                }
                break;
            case RoundingMode.HALF_EVEN:
                if (isEven)
                {
                    if (isHalf || isLess)
                    {
                        intPart.num = intPart.num.subtract(BigInteger.ONE);
                    }
                }
                else
                {
                    if (!isLess)
                    {
                        intPart.num = intPart.num.add(BigInteger.ONE);
                    }
                }
                break;

            case RoundingMode.UNNECESSARY:
                throw new ArithmeticException("Rational.round(UNNECESSARY)");
            default:
                throw new IllegalArgumentException("Rational.round(garbage)");
        }

        return isNeg ? intPart.negate() : intPart;
    }

    public BigInteger toBigInteger(RoundingMode rnd)
    {
        return round(rnd).num;
    }
    public BigInteger toBigInteger()
    {
        return toBigInteger(RoundingMode.DOWN);
    }
    @Override public int intValue()
    {
        return toBigInteger().intValue();
    }
    @Override public int longValue()
    {
        return toBigInteger().longValue();
    }
    public BigInteger ceil()
    {
        return toBigInteger(RoundingMode.CEILING);
    }
    public BigInteger floor()
    {
        return toBigInteger(RoundingMode.FLOOR);
    }
    public BigInteger rint()
    {
        return toBigInteger(RoundingMode.HALF_EVEN);
    }

    public Rational[] getParts()
    {
        Rational intPart  = new Rational(num.divide(den));
        Rational fracPart = subtract(intPart);
        return new Rational[2] { intPart, fracPart };
    }

    public Rational remainder(Rational divisor, RoundingMode rnd)
    {
        return subtract(divide(divisor).round(rnd).multiply(divisor));
    }
    public Rational IEEEremainder(Rational divisor)
    {
        return remainder(divisor, RoundingMode.HALF_EVEN);
    }

    public Rational pow(int exp)
    {
        //
        // We know we're in lowest terms.
        //
        Rational newrat = new Rational();
        newrat.num = num.pow(exp);
        newrat.den = den.pow(exp);
        return newrat;
    }
    public Rational sqr()
    {
        Rational newrat = new Rational();
        newrat.num = num.multiply(num);
        newrat.den = den.multiply(den);
        return newrat;
    }

    //
    // Two ways to approximate square roots:
    //
    public Rational sqrt()
    {
        int sig = signum();
        if (sig < 0)
        {
            throw new ArithmeticException(
                "Extracting the square root of a negative number");
        }
        if (0 == sig)
        {
            return ZERO;
        }
        return new Rational(Math.sqrt(doubleValue()));
    }
    public Rational sqrt(double accuracy)
    {
        int sig = signum();
        if (sig < 0)
        {
            throw new ArithmeticException(
                "Extracting the square root of a negative number");
        }
        if (0 == sig)
        {
            return ZERO;
        }
        // two-arg ctor validates accuracy
        return new Rational(Math.sqrt(doubleValue()), accuracy);
    }
}

// End of Rational.java
